# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module is responsible for launching local protocol executions:

1- Check if the protocol will be submitted to a queue (using Queue template from config)
2- Build the command that will be launched.

"""

import os
import re
import logging
logger = logging.getLogger(__file__)
from subprocess import Popen, PIPE
import pyworkflow as pw
from pyworkflow.exceptions import PyworkflowException
from pyworkflow.utils import (redStr, greenStr, makeFilePath, join, process,
                              getHostFullName)
from pyworkflow.protocol.constants import UNKNOWN_JOBID, STATUS_FAILED, STATUS_FINISHED, STATUS_RUNNING


# ******************************************************************
# *         Public functions provided by the module
# ******************************************************************

def launch(protocol, wait=False, stdin=None, stdout=None, stderr=None):
    """ This function should be used to launch a protocol. """
    _launchLocal(protocol, wait, stdin, stdout, stderr)


def stop(protocol):
    """
    Stop function for three scenarios:
    - If the queue is not used, kill the main protocol process and its child processes.
    - If the queue is used and the entire protocol is sent to the queue, cancel the job running the protocol using
    scancel.
    - If the queue is used and individual steps are sent to the queue, cancel all active jobs and kill the main protocol
    process and its child processes.
    """
    if protocol.useQueue() and not protocol.isScheduled():
        jobIds = protocol.getJobIds()
        for jobId in jobIds: # Iter even though it contains only one jobId
            host = protocol.getHostConfig()
            cancelCmd = host.getCancelCommand() % {'JOB_ID': jobId}
            logger.info(cancelCmd)
            _run(cancelCmd, wait=True)

        if protocol.useQueueForSteps():
            process.killWithChilds(protocol.getPid())
    else:
        process.killWithChilds(protocol.getPid())


def schedule(protocol, initialSleepTime=0, wait=False):
    """ Use this function to schedule protocols that are not ready to
    run yet. Right now it only make sense to schedule jobs locally.
    """
    cmd = '%s %s' % (pw.PYTHON, pw.getScheduleScript())
    cmd += ' "%s" "%s" %s "%s" --initial_sleep %s' % (protocol.getProject().path,
                              protocol.getDbPath(),
                              protocol.strId(),
                              protocol.getScheduleLog(),
                              initialSleepTime)
    pid = _run(cmd, wait)
    protocol.setPid(pid)  # Set correctly the pid


# ******************************************************************
# *         Internal utility functions
# ******************************************************************
def _runsLocally(protocol):
    """ Return True if this protocol is running in this machine,
    where the PID makes sense.
    """
    return protocol.getHostFullName() == getHostFullName()


# ******************************************************************
# *                 Function related to LAUNCH
# ******************************************************************
def _getAppsProgram(prog):
    """ Get a command to launch a program under the apps folder.
    """
    return "%s %s" % (pw.PYTHON, pw.join(pw.APPS, prog))


def _launchLocal(protocol, wait, stdin=None, stdout=None, stderr=None):
    """

    :param protocol: Protocol to launch
    :param wait: Pass true if you want to wait for the process to finish
    :param stdin: stdin object to direct stdin to
    :param stdout: stdout object to send process stdout
    :param stderr: stderr object to send process stderr
    """

    command = '{python} {prot_run} "{project_path}" "{db_path}" {prot_id} "{stdout_log}" "{stderr_log}"'.format(
        python=pw.PYTHON,
        prot_run=pw.join(pw.APPS, 'pw_protocol_run.py'),
        project_path=protocol.getProject().path,
        db_path=protocol.getDbPath(),
        prot_id=protocol.strId(),
        # We make them absolute in case working dir is not passed to the node when running through a queue.
        # The reason is that since 3.0.27, the first thing that is affected by the current working dir is the
        # creation of the logs. Before event than loading the project, which was and is setting the working dir to
        # the project path. IMPORTANT: This assumes the paths before the queue and after the queue (nodes) are the same
        # Which I think is safe since we are passing here "project_path" that is absolute.
        stdout_log=os.path.abspath(protocol.getStdoutLog()),
        stderr_log=os.path.abspath(protocol.getStderrLog())
    )

    hostConfig = protocol.getHostConfig()

    # Clean Pid and JobIds
    protocol.cleanExecutionAttributes()
    protocol._store(protocol._jobId)

    # Handle three use cases: one will use the job ID, and the other two will use the process ID.
    if protocol.useQueueForProtocol():  # Retrieve the job ID and set it; this will be used to control the protocol.
        submitDict = dict(hostConfig.getQueuesDefault())
        submitDict.update(protocol.getSubmitDict())
        submitDict['JOB_COMMAND'] = command
        jobId = _submit(hostConfig, submitDict)
        if jobId is None or jobId == UNKNOWN_JOBID:
            protocol.setStatus(STATUS_FAILED)
        else:
            protocol.setJobId(jobId)
            protocol.setPid(0)  # we go through the queue, so we rely on the jobId
    else:  # If not, retrieve and set the process ID (both for normal execution or when using the queue for steps)
        pId = _run(command, wait, stdin, stdout, stderr)
        protocol.setPid(pId)


def analyzeFormattingTypeError(string, dictionary):
    """ receives a string with %(VARS) to be replaced with a dictionary
     it splits te string by \n and test the formatting per line. Raises an exception if any line fails
     with all problems found"""

    # Do the replacement line by line
    lines = string.split("\n")

    problematicLines = []
    for line in lines:
        try:
            line % dictionary
        except KeyError as e:
            problematicLines.append(line + " --> variable not present in this context.")
        except Exception as e:
            problematicLines.append(line + " --> " + str(e))

    if problematicLines:
        return PyworkflowException('Following lines in %s seems to be problematic.\n'
                                   'Values known in this context are: \n%s'
                                   'Please review its format or content.\n%s' % (dictionary, pw.Config.SCIPION_HOSTS, "\n".join(problematicLines)),
                                   url=pw.DOCSITEURLS.HOST_CONFIG)


def _submit(hostConfig, submitDict, cwd=None, env=None):
    """ Submit a protocol to a queue system. Return its job id.
    """
    # Create first the submission script to be launched
    # formatting using the template
    template = hostConfig.getSubmitTemplate()

    try:
        template = template % submitDict
    except Exception as e:
        # Capture parsing errors
        exception = analyzeFormattingTypeError(template, submitDict)

        if exception:
            raise exception
        else:
            # If there is no exception, then raise actual one
            raise e

    # FIXME: CREATE THE PATH FIRST
    scripPath = submitDict['JOB_SCRIPT']
    f = open(scripPath, 'w')
    # Ensure the path exists
    makeFilePath(scripPath)
    # Add some line ends because in some clusters it fails
    # to submit jobs if the submit script does not have end of line
    f.write(template + '\n\n')
    f.close()
    # This should format the command using a template like: 
    # "qsub %(JOB_SCRIPT)s"
    command = hostConfig.getSubmitCommand() % submitDict
    gcmd = greenStr(command)
    logger.info("** Submitting to queue: '%s'" % gcmd)

    p = Popen(command, shell=True, stdout=PIPE, cwd=cwd, env=env)
    out = p.communicate()[0]
    # Try to parse the result of qsub, searching for a number (jobId)
    # Review this, seems to exclusive to torque batch system
    s = re.search(r'(\d+)', str(out))
    if p.returncode == 0 and s:
        job = int(s.group(0))
        logger.info("Launched job with id %s" % job)
        return job
    else:
        logger.info("Couldn't submit to queue for reason: %s " % redStr(out.decode()))
        return UNKNOWN_JOBID

def _checkJobStatus(hostConfig, jobid):
    """
    General method to verify the job status in the queue based on the jobId and host.conf CHECK_COMMAND
    returns: STATUS_FINISHED (finished) or STATUS_RUNNING (running)
    """
    command = hostConfig.getCheckCommand() % {"JOB_ID": jobid}
    logger.debug("checking job status for %s: %s" % (jobid, command))

    p = Popen(command, shell=True, stdout=PIPE, preexec_fn=os.setsid)

    out = p.communicate()[0].decode(errors='backslashreplace')

    jobDoneRegex = hostConfig.getJobDoneRegex()
    logger.debug("Queue engine replied %s, variable JOB_DONE_REGEX has %s" % (out, jobDoneRegex))
    # If nothing is returned we assume job is no longer in queue and thus finished
    if out == "":
        logger.warning("Empty response from queue system to job (%s)" % jobid)
        return STATUS_FINISHED

    # If some string is returned we use the JOB_DONE_REGEX variable (if present) to infer the status
    elif jobDoneRegex is not None:
        s = re.search(jobDoneRegex, out)
        if s:
            logger.debug("Job (%s) finished" % jobid)
            return STATUS_FINISHED
        else:
            logger.debug("Job (%s) still running" % jobid)
            return STATUS_RUNNING
    # If JOB_DONE_REGEX is not defined and queue has returned something we assume that job is still running
    else:
        return STATUS_RUNNING

def _run(command, wait, stdin=None, stdout=None, stderr=None):
    """ Execute a command in a subprocess and return the pid. """
    gcmd = greenStr(command)
    logger.info("** Running command: '%s'" % gcmd)
    p = Popen(command, shell=True, stdout=stdout, stderr=stderr)
    pid = p.pid
    if wait:
        p.wait()

    return pid



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module have the classes for execution of protocol steps.
The basic one will run steps, one by one, after completion.
There is one based on threads to execute steps in parallel
using different threads and the last one with MPI processes.
"""

import logging
logger = logging.getLogger(__name__)
import time
import datetime
import threading
import os

import pyworkflow.utils.process as process
from pyworkflow.utils.path import getParentFolder, removeExt
from . import constants as cts

from .launch import _submit, UNKNOWN_JOBID, _checkJobStatus


class StepExecutor:
    """ Run a list of Protocol steps. """
    def __init__(self, hostConfig, **kwargs):
        self.hostConfig = hostConfig
        self.gpuList = kwargs.get(cts.GPU_LIST, None)
        self.protocol = None

    def getGpuList(self):
        """ Return the GPU list assigned to current thread. """
        return self.gpuList

    def setProtocol(self, protocol):
        """ Set protocol to append active jobs to its jobIds. """
        self.protocol = protocol

    def runJob(self, log, programName, params,           
               numberOfMpi=1, numberOfThreads=1,
               env=None, cwd=None, executable=None):
        """ This function is a wrapper around runJob, 
        providing the host configuration. 
        """
        process.runJob(log, programName, params,
                       numberOfMpi, numberOfThreads, 
                       self.hostConfig,
                       env=env, cwd=cwd, gpuList=self.getGpuList(), executable=executable)

    def _getRunnable(self, steps, n=1):
        """ Return the n steps that are 'new' and all its
        dependencies have been finished, or None if none ready.
        """
        rs = []  # return a list of runnable steps

        for s in steps:
            if (s.getStatus() == cts.STATUS_NEW and
                    all(steps[i-1].isFinished() for i in s._prerequisites)):

                if self._isStepRunnable(s):
                    rs.append(s)
                    if len(rs) == n:
                        break
        return rs
    def _isStepRunnable(self, step):
        """ Should be implemented by inherited classes to test extra conditions """
        return True

    def _arePending(self, steps):
        """ Return True if there are pending steps (either running or waiting)
        that can be done and thus enable other steps to be executed.
        """
        return any(s.isRunning() or s.isWaiting() for s in steps)
    
    def runSteps(self, steps, 
                 stepStartedCallback, 
                 stepFinishedCallback,
                 stepsCheckCallback,
                 stepsCheckSecs=3):
        # Even if this will run the steps in a single thread
        # let's follow a similar approach than the parallel one
        # In this way we can take into account the steps graph
        # dependency and also the case when using streaming

        delta = datetime.timedelta(seconds=stepsCheckSecs)
        lastCheck = datetime.datetime.now()

        while True:
            # Get a step to run, if there is any
            runnableSteps = self._getRunnable(steps)

            if runnableSteps:
                step = runnableSteps[0]
                # We found a step to work in, so let's start a new
                # thread to do the job and book it.
                step.setRunning()
                stepStartedCallback(step)
                step.run()
                doContinue = stepFinishedCallback(step)
            
                if not doContinue:
                    break

            elif self._arePending(steps):
                # We have not found any runnable step, but still there
                # there are some running or waiting for dependencies
                # So, let's wait a bit to check if something changes
                time.sleep(0.5)
            else:
                # No steps to run, neither running or waiting
                # So, we are done, either failed or finished :)
                break

            now = datetime.datetime.now()
            if now - lastCheck > delta:
                stepsCheckCallback()
                lastCheck = now

        stepsCheckCallback()  # one last check to finalize stuff


class StepThread(threading.Thread):
    """ Thread to run Steps in parallel. """
    def __init__(self, step, lock):
        threading.Thread.__init__(self)
        self.thId = step.getObjId()
        self.step = step
        self.lock = lock

    def run(self):
        error = None
        try:
            self.step._run()  # not self.step.run() , to avoid race conditions
        except Exception as e:
            error = str(e)
            logger.error("Couldn't run the code in a thread." , exc_info=e)
        finally:
            with self.lock:
                if error is None:
                    self.step.setFinished()
                else:
                    self.step.setFailed(error)


class ThreadStepExecutor(StepExecutor):
    """ Run steps in parallel using threads. """
    def __init__(self, hostConfig, nThreads, **kwargs):
        StepExecutor.__init__(self, hostConfig, **kwargs)
        self.numberOfProcs = nThreads
        # If the gpuList was specified, we need to distribute GPUs among
        # all the threads
        self.gpuDict = {}

        self._assignGPUperNode()

    def _assignGPUperNode(self):
        # If we have GPUs
        if self.gpuList:

            nThreads = self.numberOfProcs

            # Nodes: each concurrent steps
            nodes = range(1, nThreads+1)

            # Number of GPUs
            nGpu = len(self.gpuList)

            # If more GPUs than threads
            if nGpu > nThreads:

                # Get the ratio: 2 GPUs per thread? 3 GPUs per thread?
                # 3 GPU and 2 threads is rounded to 1 (flooring)
                step = int(nGpu / nThreads)
                spare = nGpu % nThreads
                fromPos = 0
                # For each node(concurrent thread)
                for node in nodes:
                    # Store the GPUS per thread:
                    # GPUs: 0 1 2
                    # Threads 2 (step 1)
                    # Node 0 : GPU 0 1
                    # Node 1 : GPU 2

                    extraGpu = 1 if spare>0 else 0
                    toPos = fromPos + step +extraGpu
                    gpusForNode = list(self.gpuList[fromPos:toPos])

                    newGpusForNode = self.cleanVoidGPUs(gpusForNode)
                    if len(newGpusForNode) == 0:
                        logger.info("Gpu slot cancelled: all were null Gpus -> %s" % gpusForNode)
                    else:
                        logger.info("GPUs %s assigned to node %s" % (newGpusForNode, node))
                        self.gpuDict[-node] = newGpusForNode

                    fromPos = toPos
                    spare-=1

            else:
                # Expand gpuList repeating until reach nThreads items
                if nThreads > nGpu:
                    logger.warning("GPUs are no longer extended. If you want all GPUs to match threads repeat as many "
                                   "GPUs as threads.")
                    # newList = self.gpuList * (int(nThreads / nGpu) + 1)
                    # self.gpuList = newList[:nThreads]

                for index, gpu in enumerate(self.gpuList):

                    if gpu == cts.VOID_GPU:
                        logger.info("Void GPU (%s) found in the list. Skipping the slot." % cts.VOID_GPU)
                    else:
                        logger.info("GPU slot for gpu %s." % gpu)
                        # Any negative number in the key means a free gpu slot. can't be 0!
                        self.gpuDict[-index-1] = [gpu]

    def cleanVoidGPUs(self, gpuList):
        newGPUList=[]
        for gpuid in gpuList:
            if gpuid == cts.VOID_GPU:
                logger.info("Void GPU detected in %s" % gpuList)
            else:
                newGPUList.append(gpuid)
        return newGPUList

    def getGpuList(self):
        """ Return the GPU list assigned to current thread
        or empty list if not using GPUs. """

        # If the node id has assigned gpus?
        nodeId = threading.current_thread().thId
        if nodeId in self.gpuDict:
            gpus = self.gpuDict.get(nodeId)
            logger.info("Reusing GPUs (%s) slot for %s" % (gpus, nodeId))
            return gpus
        else:

            gpus = self.getFreeGpuSlot(nodeId)
            if gpus is None:
                logger.warning("Step on node %s is requesting GPUs but there isn't any available. Review configuration of threads/GPUs. Returning and empty list." % nodeId)
                return []
            else:
                return gpus
    def getFreeGpuSlot(self, stepId=None):
        """ Returns a free gpu slot available or None. If node is passed it also reserves it for that node

        :param node: node to make the reserve of Gpus
        """
        for node in self.gpuDict.keys():
            # This is a free node. Book it
            if node < 0:
                gpus = self.gpuDict[node]

                if stepId is not None:
                    self.gpuDict.pop(node)
                    self.gpuDict[stepId] = gpus
                    logger.info("GPUs %s assigned to step %s" % (gpus, stepId))
                else:
                    logger.info("Free gpu slot found at %s" % node)
                return gpus

        return None
    def freeGpusSlot(self, node):
        gpus = self.gpuDict.get(node, None)

        # Some nodes/threads do not use gpus so may not be booked and not in the dictionary
        if gpus is not None:
            self.gpuDict.pop(node)
            self.gpuDict[-node] = gpus
            logger.info("GPUs %s freed from step %s" % (gpus, node))
        else:
            logger.debug("step id %s not found in GPU slots" % node)

    def _isStepRunnable(self, step):
        """ Overwrite this method to check GPUs availability"""

        if self.gpuList and step.needsGPU() and self.getFreeGpuSlot(step.getObjId()) is None:
            logger.info("Can't run step %s. Needs gpus and there are no free gpu slots" % step)
            return False

        return True

    def runSteps(self, steps, 
                 stepStartedCallback, 
                 stepFinishedCallback,
                 stepsCheckCallback,
                 stepsCheckSecs=5):
        """
        Creates threads and synchronize the steps execution.

        :param steps: list of steps to run
        :param stepStartedCallback: callback to be called before starting any step
        :param stepFinishedCallback: callback to be run after all steps are done
        :param stepsCheckCallback: callback to check if there are new steps to add (streaming)
        :param stepsCheckSecs: seconds between stepsCheckCallback calls

        """

        delta = datetime.timedelta(seconds=stepsCheckSecs)
        lastCheck = datetime.datetime.now()

        sharedLock = threading.Lock()

        runningSteps = {}  # currently running step in each node ({node: step})
        freeNodes = list(range(1, self.numberOfProcs+1))  # available nodes to send jobs
        logger.info("Execution threads: %s" % freeNodes)
        logger.info("Running steps using %s threads. 1 thread is used for this main proccess." % self.numberOfProcs)

        while True:
            # See which of the runningSteps are not really running anymore.
            # Update them and freeNodes, and call final callback for step.
            with sharedLock:
                nodesFinished = [node for node, step in runningSteps.items()
                                 if not step.isRunning()]
            doContinue = True
            for node in nodesFinished:
                step = runningSteps.pop(node)  # remove entry from runningSteps
                freeNodes.append(node)  # the node is available now
                self.freeGpusSlot(step.getObjId())
                # Notify steps termination and check if we should continue
                doContinue = stepFinishedCallback(step)
                if not doContinue:
                    break

            if not doContinue:
                break

            anyLaunched = False
            # If there are available nodes, send next runnable step.
            with sharedLock:
                if freeNodes:
                    runnableSteps = self._getRunnable(steps, len(freeNodes))

                    for step in runnableSteps:
                        # We found a step to work in, so let's start a new
                        # thread to do the job and book it.
                        anyLaunched = True
                        step.setRunning()
                        stepStartedCallback(step)
                        node = freeNodes.pop(0)  # take an available node
                        runningSteps[node] = step
                        logger.debug("Running step %s on node %s" % (step, node))
                        t = StepThread(step, sharedLock)
                        # won't keep process up if main thread ends
                        t.daemon = True
                        t.start()
                anyPending = self._arePending(steps)

            if not anyLaunched:
                if anyPending:  # nothing running
                    time.sleep(3)
                else:
                    break  # yeah, we are done, either failed or finished :)

            now = datetime.datetime.now()
            if now - lastCheck > delta:
                stepsCheckCallback()
                lastCheck = now

        stepsCheckCallback()

        # Wait for all threads now.
        for t in threading.enumerate():
            if t is not threading.current_thread():
                t.join()


class QueueStepExecutor(ThreadStepExecutor):
    def __init__(self, hostConfig, submitDict, nThreads, **kwargs):
        ThreadStepExecutor.__init__(self, hostConfig, nThreads, **kwargs)
        self.submitDict = submitDict
        # Command counter per thread
        self.threadCommands = {}

        if nThreads > 1:
            self.runJobs = ThreadStepExecutor.runSteps
        else:
            self.runJobs = StepExecutor.runSteps

        self.renameGpuIds()

    def renameGpuIds(self):
        """ Reorganize the gpus ids starting from 0 since the queue engine is the one assigning them.
        https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#env-vars """
        for threadId, gpuList in self.gpuDict.items():
            for i in range(len(gpuList)):
                self.gpuDict[threadId][i] = i

        logger.debug("Updated gpus ids rebase starting from 0: %s per thread" %self.gpuDict)

    def getThreadJobId(self, stepId):
        """ Returns the job id extension assigned to each thread/step """
        if not stepId in self.threadCommands:
            self.threadCommands[stepId] = 0

        self.threadCommands[stepId] += 1

        return self.threadCommands[stepId]

    def runJob(self, log, programName, params, numberOfMpi=1, numberOfThreads=1, env=None, cwd=None, executable=None):
        threadId = threading.current_thread().thId
        submitDict = dict(self.hostConfig.getQueuesDefault())
        submitDict.update(self.submitDict)
        submitDict['JOB_COMMAND'] = process.buildRunCommand(programName, params, numberOfMpi,
                                                            self.hostConfig, env,
                                                            gpuList=self.getGpuList())
        threadJobId = self.getThreadJobId(threadId)
        subthreadId = '-%s-%s' % (threadId, threadJobId)
        submitDict['JOB_NAME'] = submitDict['JOB_NAME'] + subthreadId
        submitDict['JOB_SCRIPT'] = os.path.abspath(removeExt(submitDict['JOB_SCRIPT']) + subthreadId + ".job")
        submitDict['JOB_LOGS'] = os.path.join(getParentFolder(submitDict['JOB_SCRIPT']), submitDict['JOB_NAME'])

        jobid = _submit(self.hostConfig, submitDict, cwd, env)
        self.protocol.appendJobId(jobid)  # append active jobs
        self.protocol._store(self.protocol._jobId)

        if (jobid is None) or (jobid == UNKNOWN_JOBID):
            logger.info("jobId is none therefore we set it to fail")
            raise Exception("Failed to submit to queue.")

        status = cts.STATUS_RUNNING
        wait = 3

        # Check status while job running
        # REVIEW this to minimize the overhead in time put by this delay check
        while _checkJobStatus(self.hostConfig, jobid) == cts.STATUS_RUNNING:
            time.sleep(wait)
            if wait < 300:
                wait += 3

        self.protocol.removeJobId(jobid)  # After completion, remove inactive jobs.
        self.protocol._store(self.protocol._jobId)

        return status
